import copy
import random
import warnings
from collections import Counter

import joblib
import numpy
import numpy as np
import pandas
import pandas as pd
import pyod.models.base
import scipy.stats
from pyod.models.copod import COPOD
from scipy.stats import stats
from sklearn.neighbors import NearestNeighbors
from sklearn.tree import DecisionTreeClassifier
from tqdm import tqdm

from sprout.classifiers.AutoEncoder import DeepAutoEncoder, SingleAutoEncoder, SingleSparseAutoEncoder
from sprout.classifiers.Classifier import get_classifier_name
from sprout.utils.general_utils import current_ms, get_full_class_name
from sprout.utils.sprout_utils import predictions_variability


class UncertaintyCalculator:
    """
    Abstract Class for uncertainty calculators. Methods to be overridden are uncertainty_strategy_name and uncertainty_scores
    """

    def uncertainty_calculator_name(self):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        """
        pass

    def full_uncertainty_calculator_name(self):
        """
        Returns the extended name of the strategy to calculate uncertainty score (as string)
        """
        return self.uncertainty_calculator_name()

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return None

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Method to compute uncertainty score for a set of data points
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        pass


class MaxProbUncertainty(UncertaintyCalculator):
    """
    Computes uncertainty via Maximum probability assigned to a class for a given data point.
    Higher probability means high uncertainty / confidence
    """

    def __init__(self):
        """
        Constructor Method
        """
        return

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Method to compute uncertainty score for a set of data points
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        if proba_array is None:
            proba_array = classifier.predict_proba(feature_values_array)
        max_p = numpy.max(proba_array, axis=1)
        return np.asarray(max_p)

    def uncertainty_calculator_name(self):
        return 'MaxProb Calculator'


class EntropyUncertainty(UncertaintyCalculator):
    """
    Computes uncertainty via Entropy of the probability array for a given data point.
    Higher entropy means low uncertainty / confidence
    """

    def __init__(self, norm=2):
        """
        Constructor Method
        :param norm: number of classes for normalization process
        """
        norm_array = np.full(norm, 1 / norm)
        self.normalization = (-norm_array * np.log2(norm_array)).sum()

    def uncertainty_score(self, proba):
        """
        Returns the entropy for a given prediction array
        :param proba: the probability array assigned by the algorithm to the data point
        :return: entropy score in the range [0, 1]
        """
        val = np.delete(proba, np.where(proba == 0))
        p = val / val.sum()
        entropy = (-p * np.log2(p)).sum()
        return (self.normalization - entropy) / self.normalization

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Method to compute uncertainty score for a set of data points
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        uncertainty = []
        if not isinstance(feature_values_array, np.ndarray):
            feature_values_array = feature_values_array.to_numpy()
        if len(feature_values_array) == len(proba_array):
            uncertainty = [self.uncertainty_score(proba_array[i]) for i in range(0, len(proba_array))]
        else:
            print("Items of the feature set have a different cardinality wrt probabilities")
        return np.asarray(uncertainty)

    def uncertainty_calculator_name(self):
        return 'Entropy Calculator'


class NeighborsUncertainty(UncertaintyCalculator):
    """
    Computes uncertainty via Agreement with label predictions of neighbours.
    Reports both on the uncertainty and on the details for the neighbours.
    """

    def __init__(self, x_train, y_train, k, labels):
        self.x_train = x_train
        self.y_train = y_train
        try:
            self.n_neighbors = int(k)
        except:
            self.n_neighbors = 19
        self.labels = labels

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return {"n_neighbors": self.n_neighbors}

    def uncertainty_calculator_name(self):
        return 'uncertainty calculator on ' + str(self.n_neighbors) + ' Neighbors'

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Computes uncertainty by predicting the labels for the k-NN of each data point.
        uncertainty score ranges from 0 (complete disagreement) to 1 (complete agreement)
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :param classifier: the classifier used for classification
        :return: dictionary of two arrays: uncertainty and Detail
        """
        start_time = current_ms()
        #print("Starting kNN search ...")
        near_neighbors = NearestNeighbors(n_neighbors=self.n_neighbors,
                                          algorithm='kd_tree',
                                          n_jobs=-1).fit(self.x_train)
        distances, indices = near_neighbors.kneighbors(feature_values_array)
        #print("kNN Search completed in " + str(current_ms() - start_time) + " ms")
        train_proba = np.asarray(classifier.predict_proba(self.x_train))
        train_classes = numpy.argmax(train_proba, axis=1)
        if proba_array is None or proba_array.shape[0] != feature_values_array.shape[0]:
            proba_array = np.asarray(classifier.predict_proba(feature_values_array))
        predict_classes = numpy.argmax(proba_array, axis=1)
        neighbour_agreement = [0.0 for i in range(len(feature_values_array))]
        neighbour_uncertainty = [0.0 for i in range(len(feature_values_array))]
        neighbour_c = [0 for i in range(len(feature_values_array))]
        for i in range(len(feature_values_array)):
            predict_neighbours = train_classes[indices[i]]
            agreements = (predict_neighbours == predict_classes[i]).sum()
            neighbour_agreement[i] = agreements / len(predict_neighbours)
            neighbour_uncertainty[i] = predictions_variability(train_proba[indices[i]])
            neighbour_c[i] = Counter(list(map(lambda x: self.labels[x], predict_neighbours))).most_common()
        return {"agreement": neighbour_agreement, "uncertainty": neighbour_uncertainty, "Detail": neighbour_c}


class ExternalSupervisedUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that runs an external classifer and calculates its confidence in the result
    """

    def __init__(self, del_clf, x_train, y_train, norm=2, unc_measure='entropy', verbose=False):
        self.del_clf = del_clf
        if unc_measure == 'entropy':
            self.u_measure = EntropyUncertainty(norm)
        else:
            self.u_measure = MaxProbUncertainty()
            unc_measure = 'max_prob'
        self.unc_measure = unc_measure
        if x_train is not None and y_train is not None:
            if isinstance(x_train, pandas.DataFrame):
                x_train = x_train.to_numpy()
            self.del_clf.fit(x_train, y_train)
            if verbose:
                print("[ExternalSupuncertainty] Fitting of '" + get_classifier_name(del_clf) + "' Completed")
        else:
            print("[ExternalSupuncertainty] Unable to train the supervised classifier - no data available")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        joblib.dump(self.del_clf, main_folder + tag + "_del_clf.joblib", compress=9)
        return {"del_clf": get_full_class_name(self.del_clf.__class__), "unc_measure": self.unc_measure}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Method to compute uncertainty score for a set of data points
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        return self.u_measure.uncertainty_scores(feature_values_array,
                                                 self.del_clf.predict_proba(feature_values_array),
                                                 self.del_clf)

    def uncertainty_calculator_name(self):
        return 'External Supervised Calculator (' + get_classifier_name(self.del_clf) + '/' \
               + str(self.unc_measure) + ')'


class ExternalUnsupervisedUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that runs an external classifer and calculates its confidence in the result
    """

    def __init__(self, del_clf, x_train, norm=2, unc_measure='entropy', verbose=False):
        self.del_clf = del_clf
        if unc_measure == 'entropy':
            self.u_measure = EntropyUncertainty(norm)
        else:
            self.u_measure = MaxProbUncertainty()
            unc_measure = 'max_prob'
        self.unc_measure = unc_measure
        if x_train is not None:
            if isinstance(x_train, pandas.DataFrame):
                x_train = x_train.to_numpy()
            self.del_clf.fit(x_train)
            if verbose:
                print("[ExternalUnsuncertainty] Fitting of '" + get_classifier_name(del_clf) + "' Completed")
        else:
            print("[ExternalUnsuncertainty] Unable to train the supervised classifier - no data available")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        joblib.dump(self.del_clf, main_folder + tag + "_del_clf.joblib", compress=9)
        return {"del_clf": get_full_class_name(self.del_clf.__class__), "unc_measure": self.unc_measure}

    def unsupervised_predict_proba(self, test_features):
        proba = self.del_clf.predict_proba(test_features)
        pred = self.del_clf.predict(test_features)
        for i in range(len(pred)):
            min_p = min(proba[i])
            max_p = max(proba[i])
            proba[i][pred[i]] = max_p
            proba[i][1 - pred[i]] = min_p
        return proba

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Method to compute uncertainty score for a set of data points
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        if isinstance(classifier, pyod.models.base.BaseDetector):
            return self.u_measure.uncertainty_scores(feature_values_array,
                                                     self.unsupervised_predict_proba(feature_values_array),
                                                     self.del_clf)
        else:
            return self.u_measure.uncertainty_scores(feature_values_array,
                                                     self.del_clf.predict_proba(feature_values_array),
                                                     self.del_clf)

    def uncertainty_calculator_name(self):
        return 'External Unsupervised Calculator (' + get_classifier_name(self.del_clf) + '/' \
               + str(self.unc_measure) + ')'


class CombinedUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that uses another classifer and calculates a combined confidence
    It uses the main classifier plus the additional classifier to calculate an unified confidence score
    """

    def __init__(self, del_clf, x_train, y_train=None, norm=2, verbose=False):
        self.del_clf = del_clf
        self.u_measure = EntropyUncertainty(norm)
        if x_train is not None:
            if isinstance(x_train, pandas.DataFrame):
                x_train = x_train.to_numpy()
            start_time = current_ms()

            if isinstance(self.del_clf, pyod.models.base.BaseDetector):
                self.del_clf.fit(x_train)
            else:
                self.del_clf.fit(x_train, y_train)
            if verbose:
                print("[Combineduncertainty] Fitting of '" + get_classifier_name(del_clf) + "' Completed in " +
                    str(current_ms() - start_time) + " ms")
        else:
            print("[CombinedUncertainty] Unable to train combined classifier - no data available")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        joblib.dump(self.del_clf, main_folder + tag + "_del_clf.joblib", compress=9)
        return {"del_clf": get_full_class_name(self.del_clf.__class__)}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Returns the combined uncertainty calculated using the main classifier plus the additional classifier
        Score ranges from
            -1 (complete and strong disagreement between the two classifiers) -> low confidence
        to
            1, which represents the complete agreement between the two classifiers and thus high confidence
        a score of 0 represents a very uncertain prediction
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        if proba_array is None or proba_array.shape[0] != feature_values_array.shape[0]:
            pred = classifier.predict(feature_values_array)
        else:
            pred = numpy.argmax(proba_array, axis=1)
        other_pred = self.del_clf.predict(feature_values_array)
        entropy = self.u_measure.uncertainty_scores(feature_values_array, proba_array, classifier)
        other_entropy = self.u_measure.uncertainty_scores(feature_values_array,
                                                          self.del_clf.predict_proba(feature_values_array),
                                                          self.del_clf)
        return np.where(pred == other_pred, (entropy + other_entropy) / 2, -(entropy + other_entropy) / 2)

    def uncertainty_calculator_name(self):
        return 'Combined Calculator (' + get_classifier_name(self.del_clf) + ')'


class MultiCombinedUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that uses another classifer and calculates a combined confidence
    It uses the main classifier plus the additional classifier to calculate an unified confidence score
    """

    def __init__(self, clf_set, x_train, y_train=None, norm=2, verbose=False):
        self.uncertainty_set = []
        self.tag = ""
        start_time = current_ms()
        for clf in clf_set:
            self.uncertainty_set.append(CombinedUncertainty(clf, x_train, y_train, norm))
            self.tag = self.tag + get_classifier_name(clf)[0] + get_classifier_name(clf)[-1]
        self.tag = str(len(self.uncertainty_set)) + " - " + self.tag
        if verbose:
            print("[MultiCombineduncertainty] Fitting of " + str(len(clf_set)) + " classifiers completed in "
                  + str(current_ms() - start_time) + " ms")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        i = 1
        clf_names = []
        for uc in self.uncertainty_set:
            clf = uc.del_clf
            joblib.dump(clf, main_folder + tag + "_del_clf_" + str(i) + ".joblib", compress=9)
            clf_names.append(get_full_class_name(clf.__class__))
            i = i + 1
        return {"del_clfs": clf_names}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Returns the combined uncertainty averaged over many combined classifiers
        Score ranges from
            -1 (complete and strong disagreement between the two classifiers) -> low confidence
        to
            1, which represents the complete agreement between the two classifiers and thus high confidence
        a score of 0 represents a very uncertain prediction
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        multi_uncertainty = []
        for combined_uncertainty in self.uncertainty_set:
            multi_uncertainty.append(
                combined_uncertainty.uncertainty_scores(feature_values_array, proba_array, classifier))
        return numpy.average(numpy.asarray(multi_uncertainty), axis=0)

    def uncertainty_calculator_name(self):
        return 'Multiple Combined Calculator (' + str(self.tag) + ' classifiers)'


class AgreementUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that measures agreement between a set of classifiers
    It uses the main classifier plus the additional classifier to calculate an unified confidence score
    """

    def __init__(self, clf_set, x_train, y_train=None, verbose=False):
        self.clfs = []
        self.tag = ""
        start_time = current_ms()
        for clf in clf_set:
            if x_train is not None:
                if isinstance(x_train, pandas.DataFrame):
                    x_train = x_train.to_numpy()
                try:
                    if isinstance(clf, pyod.models.base.BaseDetector):
                        clf.fit(x_train)
                    else:
                        clf.fit(x_train, y_train)
                except:
                    print("Classifier '" + get_classifier_name(clf) + "' did not train correctly")
            else:
                print("Classifier '" + get_classifier_name(clf) + "' did not train correctly - no data available")
            self.clfs.append(clf)
            self.tag = self.tag + get_classifier_name(clf)[0] + get_classifier_name(clf)[-1]
        self.tag = str(len(self.clfs)) + " - " + self.tag
        if verbose:
            print("[Agreementuncertainty] Fitting of " + str(len(clf_set)) + " classifiers completed in "
                  + str(current_ms() - start_time) + " ms")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        i = 1
        clf_names = []
        for clf in self.clfs:
            joblib.dump(clf, main_folder + tag + "_clf_" + str(i) + ".joblib", compress=9)
            clf_names.append(get_full_class_name(clf.__class__))
            i = i + 1
        return {"clfs": clf_names}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Returns the combined uncertainty averaged over many combined classifiers
        Score ranges from
            0 (complete and strong disagreement between the classifiers) -> low confidence
        to
            1, which represents the complete agreement between the classifiers and thus high confidence
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        multi_uncertainty = []
        for clf_model in self.clfs:
            try:
                predictions = numpy.asarray(clf_model.predict(feature_values_array))
                multi_uncertainty.append(predictions)
            except:
                print("Classifier '" + get_classifier_name(clf_model) + "' cannot be used for prediction")
        multi_uncertainty = numpy.asarray(multi_uncertainty)
        mode_value = stats.mode(multi_uncertainty)
        scores = numpy.where(multi_uncertainty == mode_value, 1, 0)
        return numpy.average(scores, axis=1)[0]

    def uncertainty_calculator_name(self):
        return 'Agreement Calculator (' + str(self.tag) + ' classifiers)'


class ConfidenceInterval(UncertaintyCalculator):
    """
    Defines a uncertainty measure that calculates confidence intervals to derive uncertainty
    """

    def __init__(self, x_train, y_train=None, conf_level=0.9999):
        try:
            self.confidence_level = float(conf_level)
        except:
            self.confidence_level = 0.9999
        self.intervals_min = {}
        self.intervals_max = {}
        self.labels = numpy.unique(y_train)

        if x_train is not None:
            if isinstance(x_train, pandas.DataFrame):
                x_data = x_train.to_numpy()
            else:
                x_data = x_train
            if y_train is None:
                self.interval_type = 'uns'
                intervals = []
                for i in range(0, len(x_data[0])):
                    feature = x_data[:, i]
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        intervals.append(scipy.stats.t.interval(self.confidence_level,
                                                                len(feature) - 1,
                                                                loc=np.median(feature),
                                                                scale=scipy.stats.sem(feature)))
                intervals = numpy.asarray(intervals)
                self.intervals_min = numpy.asarray(intervals[:, 0])
                self.intervals_max = numpy.asarray(intervals[:, 1])
            else:
                self.interval_type = 'sup'
                for label in self.labels:
                    intervals = []
                    data = x_data[y_train == label, :]
                    for i in range(0, len(x_data[0])):
                        feature = data[:, i]
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            intervals.append(scipy.stats.t.interval(self.confidence_level,
                                                                    len(feature) - 1,
                                                                    loc=np.median(feature),
                                                                    scale=scipy.stats.sem(feature)))
                    intervals = numpy.asarray(intervals)
                    self.intervals_min[label] = numpy.asarray(intervals[:, 0])
                    self.intervals_max[label] = numpy.asarray(intervals[:, 1])
        else:
            self.interval_type = 'none'
            print("Unable to derive confidence intervals - no data available")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return {"confidence_level": self.confidence_level}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Method to compute uncertainty score for a set of data points
        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        uncertainty = []
        predicted_labels = numpy.argmax(proba_array, axis=1)
        if isinstance(feature_values_array, pandas.DataFrame):
            feature_values_array = feature_values_array.to_numpy()
        if len(feature_values_array) == len(proba_array):
            for i in range(0, len(proba_array)):
                if self.interval_type == 'sup':
                    in_left = (self.intervals_min[predicted_labels[i]] <= feature_values_array[i])
                    in_right = (feature_values_array[i] <= self.intervals_max[predicted_labels[i]])
                else:
                    in_left = (self.intervals_min <= feature_values_array[i])
                    in_right = (feature_values_array[i] <= self.intervals_max)
                uncertainty.append(numpy.average(in_left * in_right))
        else:
            print("Items of the feature set have a different cardinality wrt probabilities")
        return np.asarray(uncertainty)

    def uncertainty_calculator_name(self):
        return 'Confidence Interval (' + str(self.confidence_level) + '/' + str(self.interval_type) + ')'


class ProximityUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that creates artificial neighbours of a data point
    and checks if the classifier has a unified answer to all of those data points
    """

    def __init__(self, x_train, artificial_points: int = 10, range_wideness: float = 0.1, weighted: bool = False, verbose=False):
        self.n_artificial = int(artificial_points)
        self.range = float(range_wideness)
        self.weighted = weighted

        if x_train is not None:
            start_time = current_ms()
            if isinstance(x_train, pd.DataFrame):
                x_train = x_train.to_numpy()
            self.stds = numpy.std(x_train, axis=0)
            if verbose:
                print("Proximity Uncertainty initialized in " + str(current_ms() - start_time) + " ms")
        else:
            print("Proximity Uncertainty failed to initialize - no data available")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return {"artificial_points": self.n_artificial, "weighted": self.weighted, "range": self.range}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Returns the uncertainty after executing a given amount of simulations around the feature values
        Score ranges from -1 (likely to be misclassification) to 1 (likely to be correct classification)

        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        predicted_classes = numpy.argmax(proba_array, axis=1)
        if isinstance(feature_values_array, pd.DataFrame):
            feature_values_array = feature_values_array.to_numpy()

        # Generating MC Artificial inputs
        mc_x = []
        rng = np.random.default_rng()
        for i in range(len(feature_values_array)):
            features = feature_values_array[i]
            mc_x.extend(features + self.range * self.stds *
                        rng.uniform(low=-1, high=1, size=(self.n_artificial, len(features))))
        mc_x = np.array(mc_x)

        # Calculating predictions
        mc_predict = classifier.predict(mc_x)

        # Calculating Uncertainty
        uncertainty = []
        for i in range(len(feature_values_array)):
            mc_preds = mc_predict[i * self.n_artificial:(i + 1) * self.n_artificial]
            if self.weighted:
                artificial_features = mc_x[i * self.n_artificial:(i + 1) * self.n_artificial, :]
                distances = [numpy.linalg.norm(af - feature_values_array[i]) for af in artificial_features]
                relevance = (max(distances) - distances) + 0.1 * max(distances)
                relevance = relevance / sum(relevance)
                score = sum((mc_preds == predicted_classes[i]) * relevance)
            else:
                score = sum(mc_preds == predicted_classes[i]) / self.n_artificial
            uncertainty.append(score)

        return np.asarray(uncertainty)

    def uncertainty_calculator_name(self):
        return 'Proximity Uncertainty (' + str(self.n_artificial) + '/' + str(self.range) \
               + ('/W' if self.weighted else '') + ')'


class FeatureBaggingUncertainty(UncertaintyCalculator):
    """
    Defines a uncertainty measure that uses a Monte Carlo simulation for each class
    """

    def __init__(self, x_train, y_train, n_baggers=10, bag_type='sup'):
        self.feature_sets = []
        self.classifiers = []
        try:
            self.n_baggers = int(n_baggers)
        except:
            self.n_baggers = 10

        if x_train is not None:
            if isinstance(x_train, pandas.DataFrame):
                x_train = x_train.to_numpy()
            n_features = x_train.shape[1]

            if n_features < 20:
                bag_rate = 0.8
            elif n_features < 50:
                bag_rate = 0.7
            elif n_features < 100:
                bag_rate = 0.6
            else:
                bag_rate = 0.5
            bag_features = int(n_features * bag_rate)

            for i in tqdm(range(self.n_baggers), "Building Feature Baggers"):
                fs = random.sample(range(n_features), bag_features)
                fs.sort()
                self.feature_sets.append(fs)
                if bag_type == 'uns':
                    classifier = COPOD()
                    classifier.fit(x_train[:, fs])
                else:
                    classifier = DecisionTreeClassifier()
                    classifier.fit(x_train[:, fs], y_train)
                    bag_type = 'sup'
                self.classifiers.append(classifier)

        else:
            print("Unable to build feature baggers - no data available")

        self.bag_type = bag_type

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return {"n_baggers": self.n_baggers, "bag_type": self.bag_type}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Returns the uncertainty after executing a given amount of simulations around the feature values
        Score ranges from 0 (no agreement) to 1 (full agreement)

        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        predicted_classes = numpy.argmax(proba_array, axis=1)
        if isinstance(feature_values_array, pd.DataFrame):
            feature_values_array = feature_values_array.to_numpy()

        # Testing with all classifiers
        fs_pred = [self.classifiers[i].predict(feature_values_array[:, self.feature_sets[i]])
                   for i in range(len(self.feature_sets))]
        fs_pred = numpy.array(fs_pred).transpose()

        # Calculating Uncertainty
        uncertainty = [sum(fs_pred[i] == predicted_classes[i]) / len(fs_pred[i])
                       for i in range(len(feature_values_array))]

        return np.asarray(uncertainty)

    def uncertainty_calculator_name(self):
        return 'FeatureBagging Uncertainty (' + str(self.n_baggers) + '/' + str(self.bag_type) + ')'

#
# class ConfidenceBaggingUncertainty(CombinedUncertainty):
#     """
#     Defines a uncertainty measure that creates a bagging/boosting meta learner using a generic clf as a base estimator
#     """
#
#     def __init__(self, clf, x_train, y_train=None, n_base: int = 10, max_features: float = 0.7, sampling_ratio: float = 0.7,
#                  perc_decisors: float = None, n_decisors: int = None, n_classes: int = 2):
#         super().__init__(ConfidenceBagging(clf, n_base, max_features, sampling_ratio, perc_decisors, n_decisors),
#                          x_train, y_train, n_classes)
#         self.n_classes = n_classes
#
#     def save_params(self, main_folder, tag):
#         """
#         Returns the name of the strategy to calculate uncertainty score (as string)
#         :param main_folder: the folder where to save the details of the calculator
#         :param tag: tag to name files
#         """
#         return {"clf": self.del_clf.clf.__class__.__name__,
#                 "n_base": self.del_clf.n_base,
#                 "max_features": self.del_clf.max_features,
#                 "sampling_ratio": self.del_clf.sampling_ratio,
#                 "n_decisors": self.del_clf.n_decisors,
#                 "n_classes": self.n_classes}
#
#     def uncertainty_calculator_name(self):
#         return "ConfidenceBagger(" + str(self.del_clf.n_base) + "-" + str(self.del_clf.n_decisors) + "-" + \
#                str(self.del_clf.max_features) + "-" + str(self.del_clf.sampling_ratio) + ")"
#
#
# class ConfidenceBoostingUncertainty(CombinedUncertainty):
#     """
#     Defines a uncertainty measure that creates a bagging/boosting meta learner using a generic clf as a base estimator
#     """
#
#     def __init__(self, clf, x_train, y_train=None, n_base: int = 10, learning_rate: float = None, sampling_ratio: float = 0.5,
#                  contamination: float = None, conf_thr: float = 0.8, n_classes: int = 2):
#         super().__init__(ConfidenceBoosting(clf, n_base, learning_rate, sampling_ratio, contamination, conf_thr),
#                          x_train, y_train, n_classes)
#         self.n_classes = n_classes
#
#     def save_params(self, main_folder, tag):
#         """
#         Returns the name of the strategy to calculate uncertainty score (as string)
#         :param main_folder: the folder where to save the details of the calculator
#         :param tag: tag to name files
#         """
#         return {"clf": self.del_clf.clf.__class__.__name__,
#                 "n_base": self.del_clf.n_base,
#                 "learning_rate": self.del_clf.learning_rate,
#                 "sampling_ratio": self.del_clf.sampling_ratio,
#                 "contamination": self.del_clf.contamination,
#                 "conf_thr": self.del_clf.conf_thr,
#                 "n_classes": self.n_classes}
#
#     def uncertainty_calculator_name(self):
#         return "ConfidenceBooster(" + str(self.del_clf.n_base) + "-" + str(self.del_clf.conf_thr) + "-" + \
#                str(self.del_clf.learning_rate) + "-" + str(self.del_clf.sampling_ratio) + ")"


class ReconstructionLoss(UncertaintyCalculator):
    """
    Defines a uncertainty measure that uses the reconstruction error of an autoencoder as uncertainty measure
    """

    def __init__(self, x_train, enc_tag:str = 'simple'):
        self.ae = None
        self.enc_tag = enc_tag

        if x_train is not None:
            bottleneck = int(x_train.shape[1] / 4) if x_train.shape[1] > 8 else 2
            if enc_tag == 'deep':
                self.ae = DeepAutoEncoder(x_train.shape[1], bottleneck)
            elif enc_tag == 'sparse':
                self.ae = SingleSparseAutoEncoder(x_train.shape[1], bottleneck)
            else:
                self.enc_tag = 'simple'
                self.ae = SingleAutoEncoder(x_train.shape[1], bottleneck)
            norm_tr = copy.deepcopy(x_train)
            if isinstance(norm_tr, pandas.DataFrame):
                norm_tr = norm_tr.to_numpy()
            self.maxmin = []
            for i in range(0, norm_tr.shape[1]):
                self.maxmin.append({'min': min(norm_tr[:, i]), 'max': max(norm_tr[:, i])})
                if self.maxmin[i]['max'] - self.maxmin[i]['min'] != 0:
                    norm_tr[:, i] = (norm_tr[:, i] - self.maxmin[i]['min']) / \
                                    (self.maxmin[i]['max'] - self.maxmin[i]['min'])
            self.ae.fit(norm_tr, 50, 256, verbose=0)
        else:
            print("Unable to build autoencoder to compute loss - no data available")

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return {"enc_tag": self.enc_tag}

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Returns the uncertainty after executing a given amount of simulations around the feature values
        Score ranges from 0 (no agreement) to 1 (full agreement)

        :param classifier: the classifier used for classification
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :return: array of uncertainty scores
        """
        norm_te = copy.deepcopy(feature_values_array)
        if isinstance(norm_te, pandas.DataFrame):
            norm_te = norm_te.to_numpy()
        for i in range(0, len(self.maxmin)):
            if self.maxmin[i]['max'] - self.maxmin[i]['min'] != 0:
                norm_te[:, i] = (norm_te[:, i] - self.maxmin[i]['min']) / \
                                (self.maxmin[i]['max'] - self.maxmin[i]['min'])
        try:
            decoded_tr, loss_tr = self.ae.predict(norm_te)
        except:
            loss_tr = 0
        return np.asarray(loss_tr)

    def uncertainty_calculator_name(self):
        return 'AutoEncoder Loss (' + str(self.enc_tag) + ')'


class KNNUncertainty(UncertaintyCalculator):
    """
    Computes uncertainty via Distance wrt the k-th neighbour.
    """

    def __init__(self, x_train, k):
        self.x_train = x_train
        try:
            self.n_neighbors = int(k)
        except:
            self.n_neighbors = 19

    def save_params(self, main_folder, tag):
        """
        Returns the name of the strategy to calculate uncertainty score (as string)
        :param main_folder: the folder where to save the details of the calculator
        :param tag: tag to name files
        """
        return {"n_neighbors": self.n_neighbors}

    def uncertainty_calculator_name(self):
        return 'uncertainty calculator on ' + str(self.n_neighbors) + '-NN distance'

    def uncertainty_scores(self, feature_values_array, proba_array, classifier):
        """
        Computes uncertainty by predicting the labels for the k-NN of each data point.
        uncertainty score ranges from 0 (complete disagreement) to 1 (complete agreement)
        :param feature_values_array: the feature values of the data points in the test set
        :param proba_array: the probability arrays assigned by the algorithm to the data points
        :param classifier: the classifier used for classification
        :return: dictionary of two arrays: uncertainty and Detail
        """
        start_time = current_ms()
        #print("Starting kNN search ...")
        near_neighbors = NearestNeighbors(n_neighbors=self.n_neighbors,
                                          algorithm='kd_tree',
                                          n_jobs=-1).fit(self.x_train)
        distances, indices = near_neighbors.kneighbors(feature_values_array)
        return numpy.max(distances, axis=1)



import numpy
import pandas
import pandas as pd
import pyod
import sklearn
from autogluon.tabular import TabularPredictor
from logitboost import LogitBoost
from pyod.models.abod import ABOD
from pyod.models.base import BaseDetector
from pyod.models.cblof import CBLOF
from pyod.models.cof import COF
from pyod.models.copod import COPOD
from pyod.models.ecod import ECOD
from pyod.models.hbos import HBOS
from pyod.models.iforest import IForest
from pyod.models.inne import INNE
from pyod.models.knn import KNN
from pyod.models.loda import LODA
from pyod.models.lof import LOF
from pyod.models.mcd import MCD
from pyod.models.ocsvm import OCSVM
from pyod.models.pca import PCA
from pyod.models.so_gaal import SO_GAAL
from pyod.models.vae import VAE
from pytorch_tabnet.tab_model import TabNetClassifier
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import RandomForestClassifier, BaggingClassifier, GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, LabelEncoder
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import check_X_y
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted, check_array
from xgboost import XGBClassifier

from sprout.utils.general_utils import current_ms


# ---------------------------------- SUPPORT METHODS ------------------------------------

def get_classifier_name(clf):
    if isinstance(clf, Classifier):
        return clf.classifier_name()
    else:
        return clf.__class__.__name__


def get_feature_importance(clf):
    if isinstance(clf, Classifier):
        return clf.feature_importances()
    else:
        return clf.feature_importances_


def build_classifier(classifier, x_train, y_train, x_test, y_test, verbose=True):
    """
    Builds and Exercises a Classifier
    :param classifier: classifier object
    :param x_train: train features
    :param y_train: train label
    :param x_test: test features
    :param y_test: test label
    :param verbose: True if there should be console output
    :return: probabilities and predictions of the classifier
    """
    if verbose:
        print("\nBuilding classifier: " + get_classifier_name(classifier))

    if isinstance(x_train, pandas.DataFrame):
        train_data = x_train.to_numpy()
    else:
        train_data = x_train

    # Fitting classifier
    start_ms = current_ms()
    if isinstance(classifier, pyod.models.base.BaseDetector):
        classifier.fit(train_data)
    else:
        classifier.fit(train_data, y_train)
    train_ms = current_ms()

    # Test features have to be a numpy array
    if isinstance(x_test, pandas.DataFrame):
        test_data = x_test.to_numpy()
    else:
        test_data = x_test

    # Predicting labels
    y_pred = classifier.predict(test_data)
    test_time = current_ms() - train_ms

    # Predicting probabilities
    y_proba = classifier.predict_proba(test_data)
    if isinstance(y_proba, pd.DataFrame):
        y_proba = y_proba.to_numpy()

    if verbose:
        print(get_classifier_name(classifier) + " train/test in " + str(train_ms - start_ms) + "/" +
              str(test_time) + " ms with Accuracy: " + str(sklearn.metrics.accuracy_score(y_test, y_pred)))

    return y_proba, y_pred


def choose_classifier(clf_name, y_label, metric, contamination=None):
    """
    Returns classifier object from string description
    :param clf_name: string description of the classifier
    :param y_label: name of label column
    :param metric: target metric/loss for training
    :param contamination: used for unsupervised classifiers, fration of anomalies in train set
    :return: a classifier object
    """
    if contamination is not None and contamination > 0.5:
        contamination = 0.5
    if clf_name in {"XGB", "XGBoost"}:
        return XGBClassifier(use_label_encoder=False, eval_metric="logloss")
    elif clf_name in {"DT", "DTree", "DecisionTree"}:
        return DecisionTreeClassifier()
    elif clf_name in {"KNN", "knn", "kNN", "KNeighbours"}:
        return KNeighborsClassifier(n_neighbors=11, n_jobs=-1, algorithm="kd_tree")
    elif clf_name in {"SVM"}:
        return BaggingClassifier(SVC(gamma='auto', probability=True), max_samples=0.1, n_estimators=10)
    elif clf_name in {"LDA"}:
        return LinearDiscriminantAnalysis()
    elif clf_name in {"GNB", "GaussianNB"}:
        return Pipeline([("norm", MinMaxScaler()), ("clf", GaussianNB())])
    elif clf_name in {"MNB", "MultinomialNB"}:
        return Pipeline([("norm", MinMaxScaler()), ("clf", MultinomialNB())])
    elif clf_name in {"Regression", "LogisticRegression", "LR"}:
        return LogisticReg()
    elif clf_name in {"RF", "RandomForest"}:
        return RandomForestClassifier(n_estimators=10)
    elif clf_name in {"LB", "LogBoost", "LogitBoost"}:
        return LogitBoost(n_estimators=10)
    elif clf_name in {"TabNet", "Tabnet", "TN"}:
        return TabNet(metric="auc", verbose=0)
    elif clf_name in {"FAI", "FastAI", "FASTAI", "fastai"}:
        return FastAI(label_name=y_label, metric=metric)
    elif clf_name in {"GBC", "GradientBoosting"}:
        return GradientBoostingClassifier(n_estimators=50)
    elif clf_name in {"COPOD"}:
        return UnsupervisedClassifier(COPOD(contamination=contamination))
    elif clf_name in {"ECOD"}:
        return UnsupervisedClassifier(ECOD(contamination=contamination))
    elif clf_name in {"HBOS"}:
        return UnsupervisedClassifier(HBOS(contamination=contamination, n_bins=30))
    elif clf_name in {"MCD"}:
        return UnsupervisedClassifier(MCD(contamination=contamination, support_fraction=0.9))
    elif clf_name in {"PCA"}:
        return UnsupervisedClassifier(PCA(contamination=contamination))
    elif clf_name in {"CBLOF"}:
        return UnsupervisedClassifier(CBLOF(contamination=contamination, alpha=0.75, beta=3))
    elif clf_name in {"OCSVM", "1SVM"}:
        return UnsupervisedClassifier(OCSVM(contamination=contamination))
    elif clf_name in {"uKNN"}:
        return UnsupervisedClassifier(KNN(contamination=contamination, n_neighbors=3))
    elif clf_name in {"LOF"}:
        return UnsupervisedClassifier(LOF(contamination=contamination, n_neighbors=5))
    elif clf_name in {"INNE"}:
        return UnsupervisedClassifier(INNE(contamination=contamination))
    elif clf_name in {"FastABOD", "ABOD", "FABOD"}:
        return UnsupervisedClassifier(ABOD(contamination=contamination, method='fast', n_neighbors=7))
    elif clf_name in {"COF"}:
        return UnsupervisedClassifier(COF(contamination=contamination, n_neighbors=9))
    elif clf_name in {"IFOREST", "IForest"}:
        return UnsupervisedClassifier(IForest(contamination=contamination, n_estimators=10))
    elif clf_name in {"LODA"}:
        return UnsupervisedClassifier(LODA(contamination=contamination, n_bins=100))
    elif clf_name in {"VAE"}:
        return UnsupervisedClassifier(VAE(contamination=contamination))
    elif clf_name in {"SO_GAAL"}:
        return UnsupervisedClassifier(SO_GAAL(contamination=contamination))
    else:
        pass


def auto_bag_rate(n_features):
    """
    Method used to automatically devise a rate to include features in bagging
    :param n_features: number of features in the dataset
    :return: the rate of features to bag
    """
    if n_features < 20:
        bag_rate = 0.8
    elif n_features < 50:
        bag_rate = 0.7
    elif n_features < 100:
        bag_rate = 0.6
    else:
        bag_rate = 0.5
    return bag_rate


def predict_uns_proba(uns_clf, test_features):
    proba = uns_clf.predict_proba(test_features)
    pred = numpy.argmax(proba, axis=1)
    for i in range(len(pred)):
        min_p = min(proba[i])
        max_p = max(proba[i])
        proba[i][pred[i]] = max_p
        proba[i][1 - pred[i]] = min_p
    return proba


class Classifier(BaseEstimator, ClassifierMixin):
    """
    Basic Abstract Class for Classifiers.
    Abstract methods are only the classifier_name, with many degrees of freedom in implementing them.
    Wraps implementations from different frameworks (if needed), sklearn and many deep learning utilities
    """

    def __init__(self, clf):
        """
        Constructor of a generic Classifier
        :param clf: algorithm to be used as Classifier
        """
        self.clf = clf
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None

    def fit(self, X, y=None):

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit + other data
        if y is not None:
            self.classes_ = unique_labels(y)
        else:
            self.classes_ = [0, 1]
        #self.X_ = X
        #self.y_ = y

        # Train clf
        self.clf.fit(X, y)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def feature_importances(self):
        """
        Returns feature importances
        :return: a list
        """
        return self.feature_importances_

    def predict(self, X):
        """
        Method to compute predict of a classifier
        :return: array of predicted class
        """
        probas = self.predict_proba(X)
        return self.classes_[numpy.argmax(probas, axis=1)]

    def predict_proba(self, X):
        """
        Method to compute probabilities of predicted classes
        :return: array of probabilities for each classes
        """

        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        return self.clf.predict_proba(X)

    def predict_confidence(self, X):
        """
        Method to compute confidence in the predicted class
        :return: max probability as default
        """
        probas = self.predict_proba(X)
        return numpy.max(probas, axis=1)

    def compute_feature_importances(self):
        """
        Outputs feature ranking in building a Classifier
        :return: ndarray containing feature ranks
        """
        if hasattr(self.clf, 'feature_importances_'):
            return self.clf.feature_importances_
        elif hasattr(self.clf, 'coef_'):
            return numpy.sum(numpy.absolute(self.clf.coef_), axis=0)
        return []

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__

    def get_params(self, deep=True):
        return {'clf': self.clf}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self.clf, parameter, value)
        return self


class UnsupervisedClassifier(Classifier, BaseDetector):
    """
    Wrapper for unsupervised classifiers belonging to the library PYOD
    """

    def __init__(self, clf):
        """
        Constructor of a generic UnsupervisedClassifier. Assumes that clf is an algorithm from pyod
        :param clf: pyod algorithm to be used as Classifier
        """
        self.clf = clf
        self.contamination = clf.contamination
        self._estimator_type = "classifier"
        self.feature_importances_ = None
        self.X_ = None
        self.y_ = None

    def fit(self, X, y=None):

        # Store the classes seen during fit + other data
        self.classes_ = [0, 1]
        self.X_ = X
        self.y_ = None

        # Train clf
        self.clf.fit(X)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def decision_function(self, X):
        """
        pyod function to override. Calls the wrapped classifier.
        :param X: test set
        :return: decision function
        """
        return self.clf.decision_function(X)

    def predict_proba(self, X):
        """
        Method to compute probabilities of predicted classes.
        It has to e overridden since PYOD's implementation of predict_proba is wrong
        :return: array of probabilities for each classes
        """

        # Check if fit has been called
        check_is_fitted(self)
        X = check_array(X)

        pred_score = self.decision_function(X)
        probs = numpy.zeros((X.shape[0], 2))
        if isinstance(self.contamination, (float, int)):
            pred_thr = pred_score - self.clf.threshold_
        else:
            pred_thr = 0.5
        min_pt = min(pred_thr)
        max_pt = max(pred_thr)
        anomaly = pred_thr > 0
        cont = numpy.asarray([pred_thr[i] / max_pt if anomaly[i] else (pred_thr[i] / min_pt if min_pt != 0 else 0.2)
                              for i in range(0, len(pred_thr))])
        probs[:, 0] = 0.5 + cont / 2
        probs[:, 1] = 1 - probs[:, 0]
        probs[anomaly, 0], probs[anomaly, 1] = probs[anomaly, 1], probs[anomaly, 0]
        return probs

    def predict(self, X):
        """
        Method to compute predict of a classifier
        :return: array of predicted class
        """
        probas = self.predict_proba(X)
        return numpy.argmax(probas, axis=1)

    def classifier_name(self):
        """
        Returns the name of the classifier (as string)
        """
        return self.clf.__class__.__name__


class TabNet(Classifier):
    """
    Wrapper for the torch.tabnet algorithm
    """

    def __init__(self, metric=None, verbose=0):
        Classifier.__init__(self, TabNetClassifier(verbose=verbose))
        self.metric = metric
        self.verbose = verbose

    def fit(self, x_train, y_train):
        if isinstance(x_train, pandas.DataFrame):
            x_train = x_train.to_numpy()
        if self.metric is None:
            self.clf.fit(X_train=x_train, y_train=y_train, max_epochs=40, batch_size=1024, eval_metric=['auc'],
                         patience=2)
        else:
            self.clf.fit(X_train=x_train, y_train=y_train, max_epochs=40, batch_size=1024,
                         eval_metric=[self.metric], patience=2)
        self.classes_ = numpy.unique(y_train)
        self.feature_importances_ = self.get_feature_importances()

    def get_feature_importances(self):
        return self.clf.feature_importances_

    def predict(self, x_test):
        if isinstance(x_test, pandas.DataFrame):
            x_test = x_test.to_numpy()
        return self.clf.predict(x_test)

    def predict_proba(self, x_test):
        if isinstance(x_test, pandas.DataFrame):
            x_test = x_test.to_numpy()
        return self.clf.predict_proba(x_test)

    def classifier_name(self):
        return "TabNet"


class AutoGluon(Classifier):
    """
    Wrapper for classifiers taken from Gluon library
    clf_name options are
    ‘GBM’ (LightGBM)
    ‘CAT’ (CatBoost)
    ‘XGB’ (XGBoost)
    ‘RF’ (random forest)
    ‘XT’ (extremely randomized trees)
    ‘KNN’ (k-nearest neighbors)
    ‘LR’ (linear regression)
    ‘NN’ (neural network with MXNet backend)
    ‘FASTAI’ (neural network with FastAI backend)
    """

    def __init__(self, label_name, clf_name, metric, verbose=0):
        Classifier.__init__(self, TabularPredictor(label=label_name, eval_metric=metric, verbosity=verbose))
        self.label_name = label_name
        self.clf_name = clf_name
        self.metric = metric
        self.verbose = verbose
        self.feature_importance = []
        self.feature_names = None

    def fit(self, x_train, y_train):
        path = './AutogluonModels/' + str(current_ms())
        self.classes_ = numpy.unique(y_train)
        self.clf = TabularPredictor(label=self.label_name, eval_metric=self.metric,
                                    path=path, verbosity=self.verbose)
        if self.feature_names is None:
            self.feature_names = ['col' + str(i) for i in range(0, x_train.shape[1])]
        df = pd.DataFrame(data=x_train.copy(), columns=self.feature_names)
        df[self.label_name] = y_train
        self.clf.fit(train_data=df, hyperparameters={self.clf_name: {}})
        self.feature_importances_ = self.clf.feature_importance(df)

        self.trained = True

    def get_feature_importances(self):
        return self.feature_importances_

    def predict(self, x_test):
        df = pd.DataFrame(data=x_test, columns=self.feature_names)
        return self.clf.predict(df, as_pandas=False)

    def predict_proba(self, x_test):
        df = pd.DataFrame(data=x_test, columns=self.feature_names)
        return self.clf.predict_proba(df, as_pandas=False)

    def classifier_name(self):
        return "AutoGluon"


class FastAI(AutoGluon):
    """
    Wrapper for the gluon.FastAI algorithm
    """

    def __init__(self, label_name, metric, verbose=0):
        AutoGluon.__init__(self, label_name, "FASTAI", metric, verbose)

    def classifier_name(self):
        return "FastAI"


class GBM(AutoGluon):
    """
    Wrapper for the gluon.LightGBM algorithm
    """

    def __init__(self, label_name, metric):
        AutoGluon.__init__(self, label_name, "GBM", metric)

    def classifier_name(self):
        return "GBM"


class MXNet(AutoGluon):
    """
    Wrapper for the gluon.MXNet algorithm (to be debugged)
    """

    def __init__(self, label_name):
        AutoGluon.__init__(self, label_name, "NN")

    def classifier_name(self):
        return "MXNet"


class LogisticReg(Classifier):
    """
    Wrapper for the sklearn.LogisticRegression algorithm
    """

    def __init__(self):
        Classifier.__init__(self, LogisticRegression(solver='sag',
                                                     random_state=0,
                                                     multi_class='ovr',
                                                     max_iter=10000,
                                                     n_jobs=10,
                                                     tol=0.1))

    def classifier_name(self):
        return "LogisticRegression"


class XGB(Classifier):
    """
    Wrapper for the sklearn.LogisticRegression algorithm
    """

    def __init__(self, n_estimators=100):
        Classifier.__init__(self, XGBClassifier(n_estimators=n_estimators))
        self.l_encoder = None

    def fit(self, X, y=None):

        # Check that X and y have correct shape
        X, y = check_X_y(X, y)

        # Store the classes seen during fit + other data
        self.classes_ = unique_labels(y)
        self.l_encoder = LabelEncoder()
        y = self.l_encoder.fit_transform(y)

        #self.X_ = X
        #self.y_ = y

        # Train clf
        self.clf.fit(X, y)
        self.feature_importances_ = self.compute_feature_importances()

        # Return the classifier
        return self

    def classifier_name(self):
        return "XGBClassifier"


class SupportVectorMachine(Classifier):
    """
    Wrapper for the sklearn.SVC algorithm
    """

    def __init__(self, kernel, degree):
        Classifier.__init__(self, SVC(kernel=kernel, degree=degree, probability=True, max_iter=10000))
        self.kernel = kernel
        self.degree = degree

    def classifier_name(self):
        return "SupportVectorMachine(kernel=" + str(self.kernel) + ")"



import copy
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import keras
import numpy
import pandas
from keras import layers, regularizers


class AutoEncoder:
    """
    Class that contains the abstract methods of an AutoEncoder (Keras)
    """

    def __init__(self, input_size, min_size):
        """
        Constructor of a generic Classifier
        :param input_size: size of the input
        """
        self.trained = False
        self.autoencoder = self.init_autoencoder(input_size, min_size)

    def init_autoencoder(self, input_size, min_size):

        keras_input = keras.Input(shape=(input_size,))

        encoded, decoded = self.build_autoencoder(input_size, min_size, keras_input)

        # This model maps an input to its reconstruction
        autoencoder = keras.Model(keras_input, decoded)
        autoencoder.compile(optimizer='adam', loss='binary_crossentropy')

        return autoencoder

    def build_autoencoder(self, input_size, min_size):
        print("Warning: Method not overridden")
        pass

    def fit(self, x_train, epochs=50, batch_size=256, train_val_split=0.8, verbose=0):
        """
        Fits an AutoEncoder
        :param train_val_split: split between train and validation set
        :param batch_size: batch size
        :param epochs: number of epochs
        :param x_train: feature set
        """
        if isinstance(x_train, pandas.DataFrame):
            data_set = copy.deepcopy(x_train.values)
        else:
            data_set = copy.deepcopy(x_train)
        numpy.random.shuffle(data_set)
        split_index = int(train_val_split*data_set.shape[0])
        train_set, val_set = data_set[:split_index, :], data_set[split_index:, :]
        self.autoencoder.fit(train_set, train_set, epochs=epochs, batch_size=batch_size,
                             shuffle=True, validation_data=(val_set, val_set), verbose=verbose)
        self.trained = True

    def is_trained(self):
        """
        Flags if train was executed
        :return: True if trained, False otherwise
        """
        return self.trained

    def predict(self, x_test):
        """
        Method to compute predict of a classifier
        :return: autoencoded input
        """
        decoded_test = self.autoencoder.predict(x_test)
        mae_loss = keras.losses.mae(decoded_test, x_test).numpy()
        return decoded_test, mae_loss


class SingleAutoEncoder(AutoEncoder):

    def build_autoencoder(self, input_size, min_size, keras_input):

        # "encoded" is the encoded representation of the input
        encoded = layers.Dense(min_size, activation='relu')(keras_input)
        # "decoded" is the lossy reconstruction of the input
        decoded = layers.Dense(input_size, activation='sigmoid')(encoded)

        return encoded, decoded


class SingleSparseAutoEncoder(AutoEncoder):

    def build_autoencoder(self, input_size, min_size, keras_input):
        # Add a Dense layer with a L1 activity regularizer
        encoded = layers.Dense(min_size, activation='relu',
                               activity_regularizer=regularizers.l1(10e-5))(keras_input)
        decoded = layers.Dense(input_size, activation='sigmoid')(encoded)

        return encoded, decoded


class DeepAutoEncoder(AutoEncoder):

    def build_autoencoder(self, input_size, min_size, keras_input):
        # Add a Dense layer with a L1 activity regularizer
        step = input_size - min_size/3
        encoded = layers.Dense(min_size + 2*step, activation='relu')(keras_input)
        encoded = layers.Dense(min_size + step, activation='relu')(encoded)
        encoded = layers.Dense(min_size, activation='relu')(encoded)

        decoded = layers.Dense(min_size + step, activation='relu')(encoded)
        decoded = layers.Dense(min_size + 2*step, activation='relu')(decoded)
        decoded = layers.Dense(input_size, activation='sigmoid')(decoded)

        return encoded, decoded



import copy
import csv
import os.path

import numpy
import numpy as np
import pandas
import sklearn
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.ensemble import GradientBoostingClassifier, RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier

from sprout.classifiers.Classifier import get_classifier_name
from sprout.utils.general_utils import current_ms


def build_SPROUT_dataset(x_test, y_proba, y_pred, y_test, label_tags) -> pandas.DataFrame:
    """
    Prepares DataFrame to output SPROUT results
    :param y_proba: probabilities assigned by the classifier
    :param y_pred: predictions (classes) of the classifier
    :param y_test: labels of the test set (ground truth)
    :param label_tags: Names of the classes
    :return: a DataFrame with 4 columns
    """
    out_df = x_test.copy()
    out_df.reset_index(drop=True, inplace=True)
    out_df['true_label'] = list(map(lambda x: label_tags[x], y_test))
    out_df['predicted_label'] = list(map(lambda x: label_tags[x], y_pred))
    out_df['is_misclassification'] = np.where(out_df['true_label'] != out_df['predicted_label'], 1, 0)
    out_df['probabilities'] = [np.array2string(y_proba[i], separator=";") for i in range(len(y_proba))]
    return out_df


def read_adjudicator_calculators(model_folder: str) -> dict:
    """
    Reads Uncertainty measures associated to a specific adjudicator to load
    :param model_folder: folder where the adjudicator was saved
    :return: a dictionary
    """
    uc_dict = {}
    if os.path.exists(model_folder + "uncertainty_calculator_params.csv"):
        with open(model_folder + "uncertainty_calculator_params.csv") as csvfile:
            my_reader = csv.reader(csvfile, delimiter=',')
            next(my_reader, None)
            for row in my_reader:
                if row[0] not in uc_dict:
                    uc_dict[row[0].strip()] = {}
                uc_dict[row[0].strip()][row[1].strip()] = row[2].strip()
    return uc_dict


def predictions_variability(predictions: numpy.ndarray):
    """
    Computes the variability of predictions
    :param predictions: numpy 2d matrix of (n_predictions, n_classes)
    :return: variability/entropy of predictions
    """
    vp = 0.0
    if predictions is not None:
        stds = numpy.std(predictions, axis=0)
        vp = sum(stds)
    return vp


def compute_omission_metrics(y_true: numpy.ndarray, y_wrapper: numpy.ndarray, y_clf: numpy.ndarray, reject_tag=None) -> dict:
    """
    Assumes that y_clf may have omissions, labeled as 'reject_tag'
    :param y_true: the ground truth labels
    :param y_wrapper: the prediction of the SPROUT (wrapper) classifier
    :param y_clf: the prediction of the regular classifier
    :param reject_tag: the tag used to label rejections, default is None
    :return: a dictionary of metrics
    """
    met_dict = {}
    met_dict['alpha'] = sklearn.metrics.accuracy_score(y_true, y_clf)
    met_dict['eps'] = 1 - met_dict['alpha']
    met_dict['phi'] = numpy.count_nonzero(y_wrapper == reject_tag) / len(y_true)
    met_dict['alpha_w'] = sum(y_true == y_wrapper) / len(y_true)
    met_dict['eps_w'] = 1 - met_dict['alpha_w'] - met_dict['phi']
    met_dict['phi_c'] = sum(numpy.where((y_wrapper == reject_tag) & (y_clf == y_true), 1, 0)) / len(y_true)
    met_dict['phi_m'] = sum(numpy.where((y_wrapper == reject_tag) & (y_clf != y_true), 1, 0)) / len(y_true)
    met_dict['eps_gain'] = 0 if met_dict['eps'] == 0 else (met_dict['eps'] - met_dict['eps_w']) / met_dict['eps']
    met_dict['phi_m_ratio'] = 0 if met_dict['phi'] == 0 else met_dict['phi_m'] / met_dict['phi']
    met_dict['overall'] = 2 * met_dict['eps_gain'] * met_dict['phi_m_ratio'] / (
            met_dict['eps_gain'] + met_dict['phi_m_ratio'])
    return met_dict


# Classifiers for Detection (Binary Adjudicator). They are all supervised
CLASSIFIERS = [XGBClassifier(n_estimators=30),
               XGBClassifier(n_estimators=100),
               GradientBoostingClassifier(n_estimators=30),
               GradientBoostingClassifier(n_estimators=100),
               DecisionTreeClassifier(),
               LinearDiscriminantAnalysis(),
               RandomForestClassifier(n_estimators=30),
               RandomForestClassifier(n_estimators=100),
               GaussianNB(),
               LogisticRegression()]


def train_binary_adjudicator(x_train: numpy.ndarray, y_train: numpy.ndarray,
                             x_test: numpy.ndarray, y_test: numpy.ndarray, features: list = None,
                             misc_ratios: list = [None, 0.05, 0.1, 0.2, 0.3], verbose=True):
    """
    Trains a binary adjudicator using uncertainty measures as data, and 0/1 misclassification as label
    :param x_train: uncertainty measures of the train set
    :param y_train: 0/1 misclassification labels of the train set
    :param x_test: uncertainty measures of the validation/test set
    :param y_test: 0/1 misclassification labels of the validation/test set
    :param features: name of the features of data. Used to buld feature importance
    :param misc_ratios: list of rations for up/down-scaling misclassifications when training the adjudicator
    :param verbose: True if debug information has to be shown
    :return: the adjudicator, the feature importance, the metrics on validation/test set
    """

    # Training Binary Adjudicators to Predict Misclassifications
    best_clf = None
    best_ratio = None
    best_metrics = {"MCC": -10}

    # Formatting MISC_RATIOS
    if misc_ratios is None or len(misc_ratios) == 0:
        misc_ratios = [None]
    # Formatting features
    if features is None or len(features) <= 0:
        features = ["feature_" + str(i) for i in range(0, x_train.shape[1])]

    for clf_base in CLASSIFIERS:
        clf_name = get_classifier_name(clf_base)
        for ratio in misc_ratios:
            try:
                clf = copy.deepcopy(clf_base)
                if ratio is not None:
                    x_tr, y_tr = sample_adj_data(x_train, y_train, ratio)
                else:
                    x_tr = x_train
                    y_tr = y_train
                start_ms = current_ms()
                clf.fit(x_tr, y_tr)
                end_ms = current_ms()
                y_pred = clf.predict(x_test)
                mcc = sklearn.metrics.matthews_corrcoef(y_test, y_pred)
                if verbose:
                    print("[" + clf_name + "][ratio=" + str(ratio) + "] Accuracy: " + str(
                        sklearn.metrics.accuracy_score(y_test, y_pred))
                          + " and MCC of " + str(mcc) + " in " + str((end_ms - start_ms) / 60000) + " mins")
                if mcc > best_metrics["MCC"]:
                    best_ratio = ratio
                    best_clf = clf
                    [tn, fp], [fn, tp] = sklearn.metrics.confusion_matrix(y_test, y_pred)
                    best_metrics = {"MCC": mcc,
                                    "Accuracy": sklearn.metrics.accuracy_score(y_test, y_pred),
                                    "AUC ROC": sklearn.metrics.roc_auc_score(y_test, y_pred),
                                    "Precision": sklearn.metrics.precision_score(y_test, y_pred),
                                    "Recall": sklearn.metrics.recall_score(y_test, y_pred),
                                    "TP": tp,
                                    "TN": tn,
                                    "FP": fp,
                                    "FN": fn}
            except:
                print("Label error on classifier %s", clf_name)

    if verbose:
        print("\nBest classifier is " + get_classifier_name(best_clf) + "/" + str(best_ratio) +
              " with MCC = " + str(best_metrics["MCC"]))

    # Scores of the SPROUT wrapper
    det_dict = {"binary classifier": get_classifier_name(best_clf),
                "train data size": len(y_train),
                "train data features": numpy.asarray(features),
                "actual misclassification ratio in training set": best_ratio,
                "test_mcc": best_metrics["MCC"],
                "test_acc": best_metrics["Accuracy"],
                "test_auc": best_metrics["AUC ROC"],
                "test_p": best_metrics["Precision"],
                "test_r": best_metrics["Recall"],
                "test_tp": best_metrics["TP"],
                "test_tn": best_metrics["TN"],
                "test_fp": best_metrics["FP"],
                "test_fn": best_metrics["FN"],
                }

    if hasattr(best_clf, "feature_importances_"):
        f_imp = dict(zip(numpy.asarray(features), best_clf.feature_importances_))
    elif hasattr(best_clf, "coef_"):
        fi_scores = abs(best_clf.coef_[0])
        f_imp = dict(zip(numpy.asarray(features), fi_scores / sum(fi_scores)))
    else:
        print("No feature importance can be computed for " + get_classifier_name(best_clf))
        f_imp = {}

    return best_clf, f_imp, det_dict


def sample_adj_data(x, y, ratio: float):
    """
    Samples data to support training the binary adjudicator. Sets a given ration between 0 and 1 instances
    :param x: data
    :param y: labels
    :param ratio: ratio inbetween (0, 1)
    :return: a couple data-labels
    """
    # Creating DataFrame
    df = pandas.DataFrame(x.copy())
    df["label"] = y
    normal_frame = df.loc[df["label"] == 0]
    misc_frame = df.loc[df["label"] == 1]

    # Scaling to 'ratio'
    df_ratio = len(misc_frame.index) / len(normal_frame.index)
    if df_ratio < ratio:
        normal_frame = normal_frame.sample(frac=(df_ratio / (2 * ratio)))
    df = pandas.concat([normal_frame, misc_frame])
    df = df.sample(frac=1.0)

    return df.drop(["label"], axis=1).to_numpy(), df["label"].to_numpy()



import os
import shutil
import urllib

import numpy
import numpy as np
import pandas as pd
import sklearn as sk
from sklearn import datasets


def load_DIGITS(row_limit=np.nan, as_pandas=False):
    """
    Loads DIGITS dataset from SKLearn
    :param row_limit: int (number of data points) if you want to use a portion of the dataset
    :param as_pandas: True if output has to be a Pandas Dataframe
    :return: features and labels with train/test split, label names and feature names
    """
    return process_image_dataset("DIGITS", limit=row_limit, as_pandas=as_pandas)


def load_MNIST(row_limit=np.nan, as_pandas=False, flatten=True):
    """
    Loads MNIST dataset
    :param flatten: True if dataset should be linearized
    :param row_limit: int (number of data points) if you want to use a portion of the dataset
    :param as_pandas: True if output has to be a Pandas Dataframe
    :return: features and labels with train/test split, label names and feature names
    """
    return process_image_dataset("MNIST", limit=row_limit, as_pandas=as_pandas, flatten=flatten)


def load_FASHIONMNIST(row_limit=np.nan, as_pandas=False, flatten=True):
    """
    Loads FASHION-MNIST dataset
    :param flatten: True if dataset should be linearized
    :param row_limit: int (number of data points) if you want to use a portion of the dataset
    :param as_pandas: True if output has to be a Pandas Dataframe
    :return: features and labels with train/test split, label names and feature names
    """
    return process_image_dataset("FASHION-MNIST", limit=row_limit, as_pandas=as_pandas, flatten=flatten)


def process_image_dataset(dataset_name, limit=np.nan, as_pandas=False, flatten=True):
    """
    Gets data for analysis, provided that the dataset is an image dataset
    :param as_pandas: True if output has to be a Pandas Dataframe
    :param dataset_name: name of the image dataset
    :param limit: specifies if the number of data points has to be cropped somehow (testing purposes)
    :return: many values for analysis
    """
    if dataset_name == "DIGITS":
        mn = datasets.load_digits(as_frame=True)
        feature_list = mn.feature_names
        labels = mn.target_names
        x_digits = mn.data
        y_digits = mn.target
        if (np.isfinite(limit)) & (limit < len(y_digits)):
            x_digits = x_digits[0:limit]
            y_digits = y_digits[0:limit]
        x_tr, x_te, y_tr, y_te = sk.model_selection.train_test_split(x_digits, y_digits, test_size=0.2, shuffle=True)
        return x_tr, x_te, y_tr, y_te, labels, feature_list

    elif dataset_name == "MNIST":
        mnist_folder = "input_folder/mnist"
        if not os.path.isdir(mnist_folder):
            print("Downloading MNIST ...")
            os.makedirs(mnist_folder)
            download_file("http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz",
                          "train-images-idx3-ubyte.gz", mnist_folder)
            download_file("http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz",
                          "train-labels-idx1-ubyte.gz", mnist_folder)
            download_file("http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz",
                          "t10k-images-idx3-ubyte.gz", mnist_folder)
            download_file("http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz",
                          "t10k-labels-idx1-ubyte.gz", mnist_folder)
        return format_mnist(mnist_folder, limit, as_pandas, flatten)

    elif dataset_name == "FASHION-MNIST":
        f_mnist_folder = "input_folder/fashion"
        if not os.path.isdir(f_mnist_folder):
            print("Downloading FASHION-MNIST ...")
            os.makedirs(f_mnist_folder)
            download_file("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz",
                          "train-images-idx3-ubyte.gz", f_mnist_folder)
            download_file("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz",
                          "train-labels-idx1-ubyte.gz", f_mnist_folder)
            download_file("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz",
                          "t10k-images-idx3-ubyte.gz", f_mnist_folder)
            download_file("http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz",
                          "t10k-labels-idx1-ubyte.gz", f_mnist_folder)
        return format_mnist(f_mnist_folder, limit, as_pandas, flatten)


def download_file(file_url, file_name, folder_name):
    with urllib.request.urlopen(file_url) as response, open(folder_name + "/" + file_name, 'wb') as out_file:
        shutil.copyfileobj(response, out_file)


def process_tabular_dataset(dataset_name, label_name, limit, train_size=0.7, shuffle=True, l_encoding=True):
    """
    Method to process an input dataset as CSV
    :param limit: integer to cut dataset if needed.
    :param dataset_name: name of the file (CSV) containing the dataset
    :param label_name: name of the feature containing the label
    :return: many values for analysis
    """
    # Loading Dataset
    df = pd.read_csv(dataset_name, sep=",")

    # Shuffle
    if shuffle:
        df = df.sample(frac=1.0)
    df = df.fillna(0)
    df = df.replace('null', 0)

    # Testing Purposes
    if (np.isfinite(limit)) & (limit < len(df.index)):
        df = df[0:limit]

    if l_encoding:
        encoding = pd.factorize(df[label_name])
        y_enc = encoding[0]
        labels = encoding[1]
    else:
        y_enc = df[label_name]
        labels = numpy.unique(y_enc)

    # Basic Pre-Processing
    print("Dataset loaded: " + str(len(df.index)) + " items and " + str(len(labels)) + " labels")

    # Train/Test Split of Classifiers
    x = df.drop(columns=[label_name])
    x_no_cat = x.select_dtypes(exclude=['object'])
    feature_list = x_no_cat.columns
    x_tr, x_te, y_tr, y_te = sk.model_selection.train_test_split(x_no_cat, y_enc, test_size=1-train_size, shuffle=shuffle)

    return x_tr, x_te, y_tr, y_te, labels, feature_list


def process_binary_tabular_dataset(dataset_name, label_name, limit=numpy.nan, test_size=0.3, val_size=None):
    """
    Method to process an input dataset as CSV
    :param val_size: the size of the validation set (None if no validation)
    :param test_size: the size of the test set, default 30%
    :param limit: integer to cut dataset if needed.
    :param dataset_name: name of the file (CSV) containing the dataset
    :param label_name: name of the feature containing the label
    :return: many values for analysis
    """
    # Loading Dataset
    df = pd.read_csv(dataset_name, sep=",")

    # Shuffle
    df = df.sample(frac=1.0)
    df = df.fillna(0)
    df = df.replace('null', 0)

    # Testing Purposes
    if (np.isfinite(limit)) & (limit < len(df.index)):
        df = df[0:limit]

    print("Dataset loaded: " + str(len(df.index)) + " items")
    y_enc = numpy.where(df[label_name] == "normal", 0, 1)

    # Basic Pre-Processing
    print("Dataset loaded: " + str(len(df.index)) + " items, " + str(sum(y_enc)) + " anomalies")

    # Train/Test Split of Classifiers
    x = df.drop(columns=[label_name])
    x_no_cat = x.select_dtypes(exclude=['object'])
    feature_list = x_no_cat.columns

    # Train-Val-Test split
    x_tr, x_te, y_tr, y_te = sk.model_selection.train_test_split(x_no_cat, y_enc, test_size=test_size, shuffle=True)
    if val_size is not None:
        x_tr, x_va, y_tr, y_va = sk.model_selection.train_test_split(x_tr, y_tr, test_size=(val_size/(1-test_size)), shuffle=True)
    else:
        x_va = None
        y_va = None

    return x_tr, x_te, x_va, y_va, y_tr, y_te, ["normal", "anomaly"], feature_list


def is_image_dataset(dataset_name):
    """
    Checks if a dataset is an image dataset.
    :param dataset_name: name/path of the dataset
    :return: True if the dataset is not a tabular (CSV) dataset
    """
    return (dataset_name == "DIGITS") or (dataset_name != "MNIST") or (dataset_name != "FASHION-MNIST")


def load_mnist(path, kind='train'):
    """
    Taken from https://github.com/zalandoresearch/fashion-mnist/blob/master/utils/mnist_reader.py
    :param path: path where the mnist-like dataset is stored
    :param kind: to navigate between mnist-like archives
    :return: train/test set of the mnist-like dataset
    """
    import os
    import gzip
    import numpy as np

    """Load MNIST data from `path`"""
    labels_path = os.path.join(path,
                               '%s-labels-idx1-ubyte.gz'
                               % kind)
    images_path = os.path.join(path,
                               '%s-images-idx3-ubyte.gz'
                               % kind)

    with gzip.open(labels_path, 'rb') as lbpath:
        labels = np.frombuffer(lbpath.read(), dtype=np.uint8,
                               offset=8)

    with gzip.open(images_path, 'rb') as imgpath:
        images = np.frombuffer(imgpath.read(), dtype=np.uint8, offset=16)
        images = images.reshape(len(labels), 28, 28)

    return images, labels


def format_mnist(mnist_folder, limit, as_pandas, flatten=True):
    """
    Loads an mnist-like dataset and provides as output the train/test split plus features
    :param as_pandas: True if output has to be a Pandas Dataframe
    :param mnist_folder: folder to load the mnist-like dataset
    :param limit: specifies if the number of data points has to be cropped somehow (testing purposes)
    :return: many values for analysis
    """
    x_tr, y_tr = load_mnist(mnist_folder, kind='train')
    x_te, y_te = load_mnist(mnist_folder, kind='t10k')

    # Linearizes features in the 28x28 image
    if flatten:
        x_tr = np.stack([x.flatten() for x in x_tr])
        x_te = np.stack([x.flatten() for x in x_te])
    x_fmnist = np.concatenate([x_tr, x_te], axis=0)
    y_fmnist = np.concatenate([y_tr, y_te], axis=0)

    # Crops if needed
    if (np.isfinite(limit)) & (limit < len(x_fmnist)):
        x_tr = x_tr[0:int(limit / 2)]
        y_tr = y_tr[0:int(limit / 2)]
        x_te = x_te[0:int(limit / 2)]
        y_te = y_te[0:int(limit / 2)]

    # Lists feature names and labels
    feature_list = ["pixel_" + str(i) for i in np.arange(0, len(x_fmnist[0]), 1)]
    labels = pd.Index(np.unique(y_fmnist), dtype=object)

    if as_pandas:
        return pd.DataFrame(data=x_tr, columns=feature_list),\
               pd.DataFrame(data=x_te, columns=feature_list),\
               pd.DataFrame(data=y_tr, columns=['label']),\
               pd.DataFrame(data=y_te, columns=['label']), labels, feature_list
    else:
        return x_tr, x_te, y_tr, y_te, labels, feature_list



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
     name='sprout-ml',
     version='0.70',
     scripts=[],
     author="Tommaso Zoppi",
     author_email="tommaso.zoppi@unifi.it",
     description="SPROUT - a Safety wraPper thROugh ensembles of UncertainTy measures",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/tommyippoz/SPROUT",
     keywords=['machine learning', 'safety wrapper', 'safety monitor', 'uncertainty measures', 'ensemble'],
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )


import copy
import fnmatch
import os.path

import joblib
import numpy
import pandas
import pandas as pd
from pyod.models.base import BaseDetector
from sklearn.naive_bayes import GaussianNB

from sprout.UncertaintyCalculator import EntropyUncertainty, ConfidenceInterval, ExternalSupervisedUncertainty, \
    CombinedUncertainty, MultiCombinedUncertainty, NeighborsUncertainty, ProximityUncertainty, \
    ReconstructionLoss, ExternalUnsupervisedUncertainty, MaxProbUncertainty, AgreementUncertainty, KNNUncertainty
from sprout.classifiers.Classifier import get_classifier_name
from sprout.utils import general_utils
from sprout.utils.general_utils import get_full_class_name, current_ms
from sprout.utils.sprout_utils import read_adjudicator_calculators, compute_omission_metrics, train_binary_adjudicator


def exercise_wrapper(model_data, models_folder: str, classifier, x_train, y_train,
                     x_val, y_val, x_test, y_test, label_names: list, verbose=True) -> dict:
    """
    Exercises an experiment that takes a classifier, exercises it alone and using a SPROUTObject
    The object is either loaded (if model tag is a string pointing to an existing model) or
    created from scratch (if model is a SPROUTObject)
    :param model_data: the SPROUTObject (if from scratch) or a tag for loading existing object
    :param models_folder: folder where existing models are
    :param classifier: the main classifier
    :param x_train: train data
    :param y_train: train labels
    :param x_val: validation data (can be None)
    :param y_val: validation labels (can be None)
    :param x_test: test data
    :param y_test: test labels
    :param label_names: name of the classes of the problem (possible labels as list)
    :param verbose: True if debug information has to be shown
    :return: a dictionary with metrics
    """
    # Creating classifier clf
    if verbose:
        print("\nBuilding classifier: " + get_classifier_name(classifier))
    start_ms = current_ms()
    classifier.fit(x_train, y_train)
    train_ms = current_ms()
    y_pred = classifier.predict(x_test)
    test_time = current_ms() - train_ms
    train_time = train_ms - start_ms

    # Loading SPROUT object with a specific tag amongst those existing, or training from scratch
    if not isinstance(model_data, SPROUTObject):
        sprout_obj = SPROUTObject(models_folder=models_folder)
        sprout_obj.load_model(model_tag=model_data, clf=classifier,
                              x_train=x_train, y_train=y_train, label_names=label_names)
    else:
        sprout_obj = copy.deepcopy(model_data)
        sprout_obj.train_model(classifier, x_train, y_train, x_val, y_val)
    start_ms = current_ms()
    sprout_df, sprout_pred = sprout_obj.exercise(x=x_test, y=y_test, classifier=classifier, verbose=verbose)
    sprout_time = current_ms() - start_ms

    # Computing metrics and printing results
    metrics = compute_omission_metrics(y_test, sprout_pred, y_pred)
    metrics['train_time'] = train_time
    metrics['test_time'] = test_time
    metrics['sprout_time'] = sprout_time
    metrics['clf_name'] = get_classifier_name(classifier)
    metrics['sprout_tag'] = model_data

    return metrics


class SPROUTObject:

    def __init__(self, models_folder):
        """
        Constructor for the SPROUT object
        """
        self.trust_calculators = []
        self.models_folder = models_folder
        self.binary_adjudicator = None

    def compute_data_trust(self, data_point, classifier, verbose=False, as_pandas=True):
        """
        Outputs an array / DataFrame containing trust measures for a data point
        :param data_point: test data point
        :param classifier: classifier to evaluate trust of
        :param verbose: False if you want to suppress debug information
        :param as_pandas: True if outputs has to e a Dataframe, False if ndarray
        :return:
        """
        item = data_point.reshape(-1, 1).transpose()
        return self.compute_set_trust(item, classifier, classifier.predict_prob(item), verbose, as_pandas)

    def compute_set_trust(self, data_set, classifier, y_proba=None, verbose=True, as_pandas=True):
        """
        Outputs an array / DataFrame containing trust measures for each data point in the dataset
        :param data_set: test dataset
        :param y_proba: probabilities calculated by the classifier to evaluate trust of
        :param classifier: classifier to evaluate trust of
        :param verbose: False if you want to suppress debug information
        :param as_pandas: True if outputs has to e a Dataframe, False if ndarray
        :return:
        """
        out_df = pd.DataFrame()
        if y_proba is None:
            y_proba = classifier.predict_proba(data_set)
        if isinstance(data_set, pandas.DataFrame):
            data_set = data_set.to_numpy()
        data_set = numpy.nan_to_num(data_set, nan=0, posinf=0, neginf=0)
        for calculator in self.trust_calculators:
            if verbose:
                print("Calculating Trust Strategy: " + calculator.uncertainty_calculator_name())
            start_ms = general_utils.current_ms()
            trust_scores = calculator.uncertainty_scores(data_set, y_proba, classifier)
            trust_scores = numpy.nan_to_num(trust_scores, nan=-10, posinf=-10, neginf=-10)
            if type(trust_scores) is dict:
                for key in trust_scores:
                    out_df[calculator.uncertainty_calculator_name() + "_" + str(key)] = trust_scores[key]
            else:
                out_df[calculator.uncertainty_calculator_name()] = trust_scores
            if verbose:
                print("Completed in " + str(general_utils.current_ms() - start_ms) + " ms for " + str(len(data_set)) +
                      " items, " + str((general_utils.current_ms() - start_ms) / len(data_set)) + " ms per item")

        # Chooses output format
        if as_pandas:
            return out_df
        else:
            return out_df.to_numpy()

    def add_calculator_confidence(self, x_train, y_train=None, confidence_level=0.9999):
        """
        Confidence Interval Calculator (CM1 from paper)
        :param x_train: features in the train set
        :param y_train: labels in the train set
        :param confidence_level: size of the confidence interval (default: 0.9999)
        """
        self.trust_calculators.append(
            ConfidenceInterval(x_train=x_train,
                               y_train=y_train,
                               conf_level=confidence_level))

    def add_calculator_entropy(self, n_classes):
        """
        Entropy Calculator (CM2 from paper)
        :param n_classes: number of classes of the label
        """
        self.trust_calculators.append(EntropyUncertainty(norm=n_classes))

    def add_calculator_maxprob(self):
        """
        MaxProb Calculator
        """
        self.trust_calculators.append(MaxProbUncertainty())

    def add_calculator_bayes(self, x_train, y_train, n_classes):
        """
        External Trust Calculator using Bayes (CM3 in the paper)
        :param x_train: features in the train set
        :param y_train: labels in the train set
        :param n_classes: number of classes of the label
        """
        self.add_calculator_external(classifier=GaussianNB(), x_train=x_train, y_train=y_train, n_classes=n_classes)

    def add_calculator_external(self, classifier, n_classes, x_train, y_train=None):
        """
        External Trust Calculator (CM3 in the paper)
        :param classifier: classifier to be used as del_clf
        :param x_train: features in the train set
        :param y_train: labels in the train set (if needed)
        :param n_classes: number of classes of the label
        """
        if isinstance(classifier, BaseDetector):
            self.trust_calculators.append(
                ExternalUnsupervisedUncertainty(del_clf=classifier, x_train=x_train, norm=n_classes))
        else:
            self.trust_calculators.append(
                ExternalSupervisedUncertainty(del_clf=classifier, x_train=x_train, y_train=y_train, norm=n_classes))

    def add_calculator_combined(self, classifier, x_train, y_train, n_classes):
        """
        Combined Trust Calculator (CM4 in the paper)
        :param classifier: classifier to be used as del_clf
        :param x_train: features in the train set
        :param y_train: labels in the train set
        :param n_classes: number of classes of the label
        """
        self.trust_calculators.append(
            CombinedUncertainty(del_clf=classifier, x_train=x_train, y_train=y_train, norm=n_classes))

    def add_calculator_multicombined(self, clf_set, x_train, y_train, n_classes):
        """
        Multi-Combined Trust Calculator (CM5 in the paper)
        :param clf_set: classifiers (array) to be used as del_clf
        :param x_train: features in the train set
        :param y_train: labels in the train set
        :param n_classes: number of classes of the label
        """
        self.trust_calculators.append(
            MultiCombinedUncertainty(clf_set=clf_set, x_train=x_train, y_train=y_train, norm=n_classes))

    def add_calculator_neighbour(self, x_train, y_train, label_names, k=19):
        """
        Neighbour-based Trust Calculator (CM6 in the paper)
        :param x_train: features in the train set
        :param y_train: labels in the train set
        :param label_names: unique names in the label
        :param k: k parameter for kNN search
        """
        self.trust_calculators.append(NeighborsUncertainty(x_train=x_train, y_train=y_train, k=k, labels=label_names))

    def add_calculator_knn_distance(self, x_train, k=19):
        """
        Neighbour-based Trust Calculator (CM6x in the paper)
        :param x_train: features in the train set
        :param k: k parameter for kNN search
        """
        self.trust_calculators.append(KNNUncertainty(x_train=x_train, k=k))

    def add_calculator_agreement(self, clf_set, x_train, y_train=None):
        self.trust_calculators.append(AgreementUncertainty(clf_set=clf_set, x_train=x_train, y_train=y_train))

    def add_calculator_proximity(self, x_train, n_iterations=10, range=0.1, weighted=False):
        self.trust_calculators.append(ProximityUncertainty(x_train, n_iterations, range, weighted))

    # def add_calculator_bagging(self, base_clf, x_train, y_train, n_base: int = 10, max_features: float = 0.7,
    #                            sampling_ratio: float = 0.7, perc_decisors: float = None, n_decisors: int = None,
    #                            n_classes=2):
    #     self.trust_calculators.append(
    #         ConfidenceBaggingUncertainty(base_clf, x_train, y_train, n_base, max_features, sampling_ratio,
    #                                      perc_decisors, n_decisors, n_classes))
    #
    # def add_calculator_boosting(self, base_clf, x_train, y_train, n_base: int = 10, learning_rate: float = None,
    #                             sampling_ratio: float = 0.5, contamination: float = None, conf_thr: float = 0.8,
    #                             n_classes=2):
    #     self.trust_calculators.append(
    #         ConfidenceBoostingUncertainty(base_clf, x_train, y_train, n_base, learning_rate, sampling_ratio,
    #                                       contamination, conf_thr, n_classes))

    def add_calculator_recloss(self, x_train, tag='simple'):
        """
        External Trust Calculator using Bayes (CM3 in the paper)
        :param x_train: features in the train set
        :param tag: tagstring to initialize autoencoder
        """
        self.trust_calculators.append(ReconstructionLoss(x_train=x_train, enc_tag=tag))

    def train_model(self, main_clf, x_train, y_train, x_val, y_val, features: list = None):
        """
        Trains the binary adjudicator from scratch
        :param main_clf: the main classifier to be wrapped with the SPROUTObject
        :param x_train: train data for uncertainty measures
        :param y_train: train label for uncertainty measures
        :param x_val: validation/test data for binary adjudicator
        :param y_val: validation/test labels for binary adjudicator
        :param features: the names of features of the original dataset (used for feature importance)
        :return:
        """
        if features is None and isinstance(x_train, pandas.DataFrame):
            features = x_train.columns.to_numpy()
        unc_train = self.compute_set_trust(data_set=x_train, classifier=main_clf, verbose=False)
        unc_train = unc_train.select_dtypes(exclude=['object']).to_numpy()
        unc_val = self.compute_set_trust(data_set=x_val, classifier=main_clf, verbose=False)
        unc_val = unc_val.select_dtypes(exclude=['object']).to_numpy()
        misc_train = numpy.asarray(1 * (main_clf.predict(x_train) != y_train))
        misc_val = numpy.asarray(1 * (main_clf.predict(x_val) != y_val))
        self.binary_adjudicator, f_imp, ba_metrics = \
            train_binary_adjudicator(unc_train, misc_train, unc_val, misc_val, features, verbose=False)

        return self.binary_adjudicator

    def load_model(self, model_tag: str, clf, x_train, y_train=None, label_names: list = [0, 1]):
        """
        Loads an existing object (adjudicator + uncertainty calculators) specified by a tag.
        The tag has to match the name of a subfolder of the models_folder.
        :param model_tag: name of the object to load
        :param clf: base classifier to be wrapped in the SPROUTObject
        :param x_train: train data for uncertainty calculators
        :param y_train: train labels for uncertainty calculators
        :param label_names: names of the labels
        :return:
        """
        if os.path.exists(self.models_folder):
            if model_tag in self.get_available_models():
                model_folder = self.models_folder + str(model_tag) + "/"
                self.binary_adjudicator = joblib.load(model_folder + "binary_adj_model.joblib")
                print("Loaded Binary Adjudicator '%s' for wrapper '%s'" %
                      (get_classifier_name(self.binary_adjudicator), model_tag))
                u_calcs = read_adjudicator_calculators(model_folder)
                self.trust_calculators = []
                for uc_tag in u_calcs:
                    params = u_calcs[uc_tag]
                    calculator_name = params["calculator_class"]
                    if "Entropy" in calculator_name:
                        calc = EntropyUncertainty(norm=len(label_names))
                    elif "MaxProb" in calculator_name:
                        calc = MaxProbUncertainty()
                    elif "Neighbors" in calculator_name:
                        calc = NeighborsUncertainty(x_train=x_train, y_train=y_train,
                                                    k=params["n_neighbors"], labels=label_names)
                    elif "ExternalSupervised" in calculator_name:
                        del_clf = joblib.load(model_folder + uc_tag + "_del_clf.joblib")
                        calc = ExternalSupervisedUncertainty(del_clf=del_clf, x_train=x_train, y_train=y_train,
                                                             norm=len(label_names))
                    elif "ExternalUnsupervised" in calculator_name:
                        del_clf = joblib.load(model_folder + uc_tag + "_del_clf.joblib")
                        calc = ExternalUnsupervisedUncertainty(del_clf=del_clf, x_train=x_train,
                                                               norm=len(label_names))
                    elif ".CombinedUncertainty" in calculator_name:
                        del_clf = joblib.load(model_folder + uc_tag + "_del_clf.joblib")
                        calc = CombinedUncertainty(del_clf=del_clf, x_train=x_train, y_train=y_train,
                                                   norm=len(label_names))
                    elif "MultiCombinedUncertainty" in calculator_name:
                        del_clfs = []
                        clf_files = fnmatch.filter(os.listdir(model_folder), uc_tag + '*.joblib')
                        clf_files.sort(reverse=False)
                        for clf_name in clf_files:
                            del_clf = joblib.load(model_folder + clf_name)
                            del_clfs.append(del_clf)
                        calc = MultiCombinedUncertainty(clf_set=del_clfs, x_train=x_train, y_train=y_train,
                                                        norm=len(label_names))
                    elif "AgreementUncertainty" in calculator_name:
                        del_clfs = []
                        clf_files = fnmatch.filter(os.listdir(model_folder), uc_tag + '*.joblib')
                        clf_files.sort(reverse=False)
                        for clf_name in clf_files:
                            del_clf = joblib.load(model_folder + clf_name)
                            del_clfs.append(del_clf)
                        calc = AgreementUncertainty(clf_set=del_clfs, x_train=x_train)
                    elif "ConfidenceInterval" in calculator_name:
                        calc = ConfidenceInterval(conf_level=params["confidence_level"],
                                                  x_train=x_train, y_train=y_train)
                    elif "ProximityUncertainty" in calculator_name:
                        calc = ProximityUncertainty(x_train=x_train, artificial_points=params["artificial_points"],
                                                    range_wideness=params["range"], weighted=params["weighted"])
                    elif "ConfidenceBagging" in calculator_name:
                        calc = ConfidenceBaggingUncertainty(clf=clf, x_train=x_train, y_train=y_train,
                                                            n_base=int(params["n_base"]) if params[
                                                                                                "n_base"] != 'None' else None,
                                                            max_features=float(params["max_features"]) if params[
                                                                                                              "max_features"] != 'None' else None,
                                                            sampling_ratio=float(params["sampling_ratio"]) if
                                                            params["sampling_ratio"] != 'None' else None,
                                                            n_decisors=int(params["n_decisors"]) if params[
                                                                                                        "n_decisors"] != 'None' else None,
                                                            n_classes=len(label_names))
                    elif "ConfidenceBoosting" in calculator_name:
                        calc = ConfidenceBoostingUncertainty(clf=clf, x_train=x_train, y_train=y_train,
                                                             n_base=int(params["n_base"]) if params[
                                                                                                 "n_base"] != 'None' else None,
                                                             learning_rate=float(params["learning_rate"]) if params[
                                                                                                                 "learning_rate"] != 'None' else None,
                                                             sampling_ratio=float(params["sampling_ratio"]) if
                                                             params["sampling_ratio"] != 'None' else None,
                                                             contamination=float(params["contamination"]) if params[
                                                                                                                 "contamination"] != 'None' else None,
                                                             conf_thr=float(params["conf_thr"]) if params[
                                                                                                       "conf_thr"] != 'None' else None,
                                                             n_classes=len(label_names))
                    elif "ReconstructionLoss" in calculator_name:
                        calc = ReconstructionLoss(x_train=x_train, enc_tag=params["enc_tag"])
                    else:
                        calc = None
                    if calc is not None:
                        self.trust_calculators.append(calc)

            else:
                print("Model '" + str(model_tag) + "' does not exist")
        else:
            print("Models folder '" + self.models_folder + "' does not exist")

        return self.binary_adjudicator

    def get_available_models(self):
        """
        Returns the models available in the SPROUT repository
        :return: list of strings
        """
        return [f.name for f in os.scandir(self.models_folder) if f.is_dir()]

    def save_uncertainty_calculators(self, obj_folder: str):
        """
        Saves calculators of the SPROUTObject
        :param obj_folder: the folder in which to save the object.
        :return: nothing
        """
        # Save general info about calculators into a unique file
        with open(obj_folder + "uncertainty_calculators.txt", 'w') as f:
            f.write('# File that lists the calculators used to build this SPROUT object\n')
            for i in range(0, len(self.trust_calculators)):
                f.write('%s: %s\n' % (str(i + 1), self.trust_calculators[i].full_uncertainty_calculator_name()))

        # Saving a file for each uncertainty calculator
        pd = {}
        for i in range(0, len(self.trust_calculators)):
            params_dict = self.trust_calculators[i].save_params(obj_folder + "/",
                                                                "uncertainty_calculator_" + str(i + 1))
            if params_dict is None:
                params_dict = {}
            params_dict["calculator_class"] = get_full_class_name(self.trust_calculators[i].__class__)
            pd["uncertainty_calculator_" + str(i + 1)] = params_dict
        with open(obj_folder + "uncertainty_calculator_params.csv", 'w') as f:
            f.write('uncertainty_calculator,param_name,param_value\n')
            for u_calc in pd:
                params_dict = pd[u_calc]
                for key, value in params_dict.items():
                    f.write('%s,%s,%s\n' % (str(u_calc), str(key), str(value)))

    def exercise(self, x, y, classifier, verbose=True):
        """
        Exercised the SPROUT wrapper on a test set (x,y). Assumes that the binary adjudicator is ready,
        meaning that the object has already been loaded successfully
        :param x: the test features x_test
        :param y: the test label y_test
        :param classifier: the classifier object
        :param verbose: True if debug information has to be shown
        :return: the pandas.DataFrame containing all data plus an array of predictions of the wrapper
        """
        if isinstance(x, pandas.DataFrame):
            out_df = x.copy()
        else:
            out_df = pandas.DataFrame(data=x)
        out_df.reset_index(drop=True, inplace=True)
        out_df['true_label'] = y
        clf_pred = classifier.predict(x)
        out_df['predicted_label'] = clf_pred
        out_df['is_misclassification'] = numpy.where(out_df['true_label'] != out_df['predicted_label'], 1, 0)
        y_proba = classifier.predict_proba(x)
        out_df['probabilities'] = [numpy.array2string(y_proba[i], separator=";") for i in range(len(y_proba))]

        if self.binary_adjudicator is not None:
            # Calculating Trust Measures with SPROUT
            sp_df = self.compute_set_trust(data_set=x, classifier=classifier, verbose=verbose)
            sp_df = sp_df.select_dtypes(exclude=['object'])
            sp_df.reset_index(drop=True, inplace=True)
            out_df = pd.concat([out_df, sp_df], axis=1)

            # Predict misclassifications with SPROUT
            x_test = sp_df.select_dtypes(exclude=['object']).to_numpy()
            x_test = numpy.nan_to_num(x_test)
            misc_pred = self.binary_adjudicator.predict(x_test)
            out_df["misc_pred"] = misc_pred
            sprout_pred = numpy.where(misc_pred == 0, clf_pred, None)
            out_df["sprout_pred"] = sprout_pred

            return out_df, sprout_pred

        else:
            print('Unable to load SPROUT model')
            return out_df, clf_pred

    def predict_misclassifications(self, x, classifier, verbose=True) -> numpy.ndarray:
        """
        Predicts on a test set
        :param x: the test features x_test
        :param classifier: the classifier object
        :param verbose: True if debug information has to be shown
        :return: a ndarray
        """
        if self.binary_adjudicator is not None:
            # Calculating Trust Measures with SPROUT
            unc_m = self.compute_set_trust(data_set=x, classifier=classifier, verbose=verbose)
            x_test = unc_m.select_dtypes(exclude=['object']).to_numpy()
            x_test = numpy.nan_to_num(x_test)
            misc_pred = self.binary_adjudicator.predict(x_test)
            return misc_pred
        else:
            print("Adjudicator is not trained")
            return None
